
//  ASL THREAD LIBRARY 
//
//  ASL (Abstract Schema Language) System Execution Kernel
//
//
//
//  FILE NAME   : aslthr.c
//  MODULE TYPE : ASL Thread Library File
//  STATUS      : In development 
//  PURPOSE     : Schemas execution on a multithread environment
//  
//
//  Instituto Tecnologico Autonomo de Mexico 
//  (c) Copyright 1995-1998
//
//
//
//  HISTORY : Version (Date-Name) Modification, description. 
//
//            1.0 (Nov 20, 1995 - Claudia Calderas & Salvador Marmol), 
//            Start of development for distributed execution.

//#include <asl/aslthr.h>
//#include <asl/aslio.h>

#include "nsl_all_kernel_include.h"
#include "asl_all_rts_include.h"

// methods of class AslThreadManager

AslThreadManager::AslThreadManager(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Initializing thread execution"<<aslend;
		
	tidCount = 1;
	stackSize = NULL;
}

AslThreadManager::~AslThreadManager(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Destroying thread Manager"<<aslend;
}

AslThread *AslThreadManager::AslGetThread(int tid)
{	
	AslThread *thread;

	if (asl_debug)
		aslio<<"ASL THREAD MANAGER: Getting thread "<<tid<<" as actual"<<aslend;
		
	thrmutex.AslLock();
		
	threadList.startOnTop();
	while ((thread=threadList.getActual())!=NULL && (thread->AslGetThreadId()!=tid))
		threadList.goNext();

	thrmutex.AslUnlock();

	if (thread==NULL)
	{
		aslerr<<"ASL THREAD MANAGER ERROR: Can't get thread "<<tid<<aslend;
		return NULL;
	}
		
	return thread;
}

void AslThreadManager::AslSetStackSize(asl_size_t s)
{
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Setting stack size to "<<s<<aslend;

	stackSize = s;	
}

asl_size_t AslThreadManager::AslGetStackSize(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Getting stack size"<<aslend;

	return stackSize;
}

void AslThreadManager::AslSetConcurrency(int s)
{
	int res;
	
	if(asl_debug)
		aslio<<"ASL THREAD: Setting concurrency level to "<<s<<aslend;

	if ((res = thr_setconcurrency(s)) != 0 )
		aslerr<<"ASL THREAD ERROR: Can't set concurrency level to " << s <<aslend;
}

int AslThreadManager::AslGetConcurrency(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Getting concurrency level"<<aslend;

	return thr_getconcurrency();
}

int AslThreadManager::AslCreateThread(asl_thread_func_t f, asl_any_t t)
{
	AslThread *temp;
	
	thrmutex.AslLock();
	
	int tid = tidCount;
	
	temp = new AslThread(this, tid);
	temp->AslExecuteThread(f, t);

	threadList.insert(temp);
	
	tidCount++;
	
	thrmutex.AslUnlock();	
	return tid;
}

int AslThreadManager::AslCreateThreadLWP(asl_thread_func_t f, asl_any_t t)
{
	AslThread *temp;
	
	thrmutex.AslLock();
	
	int tid = tidCount;
	
	temp = new AslThread(this, tid);
	temp->AslRunModeLWP();
	temp->AslExecuteThread(f, t);

	threadList.insert(temp);
	
	tidCount++;

	thrmutex.AslUnlock();

	return tid;
}

int AslThreadManager::AslDeleteThread(int tid)
{
	AslThread *tempNode;

	if (asl_debug)
		aslio<<"ASL THREAD MANAGER: Deleting thread "<<tid<<aslend;
	
	tempNode = AslGetThread(tid);	
	
	if (tempNode!=NULL)
	{
		thrmutex.AslLock();
		tempNode=threadList.removeThis();
		thrmutex.AslUnlock();
	}
	else
	{
		aslerr<<"ASL THREAD MANAGER ERROR: Can't delete thread "<<tid<<aslend;
		return 0;
	}
	
	delete tempNode;

	return 1;
}

int AslThreadManager::AslGetThreadAnswer(int tid, asl_any_t* arg)
{
	AslThread *temp;

	if (asl_debug)
		aslio<<"ASL THREAD MANAGER: Geting thread answer "<<tid<<aslend;
	
	temp = AslGetThread(tid);

	return temp->AslGetThreadAnswer(arg);
}


void AslThreadManager::AslStopThread(int tid)
{
	AslThread *temp;

	if (asl_debug)
		aslio<<"ASL THREAD MANAGER: Stoping thread "<<tid<<aslend;
	
	temp = AslGetThread(tid);

	temp->AslSuspendThread();
}

void AslThreadManager::AslContinueThread(int tid)
{
	AslThread *temp;

	if (asl_debug)
		aslio<<"ASL THREAD MANAGER: Continuing thread "<<tid<<aslend;
	
	temp = AslGetThread(tid);

	temp->AslContinueThread();
}

void AslThreadManager::AslYieldProcessor(void)
{
	thr_yield();
}

void AslThreadManager::AslThreadExit(any_t r)
{
	if(asl_debug)
		aslio<<"ASL THREAD: Exiting thread"<<aslend;

	thr_exit(r);	
}

int AslThreadManager::AslKillThread(int tid)
{
	AslThread *t;
	int res;
	
	if(asl_debug)
		aslio<<"ASL THREAD MANAGER: Killing thread"<<aslend;

	t = AslGetThread(tid);
	
	if (t==NULL)
	{
		aslerr<<"ASL THREAD MANAGER ERROR: Can't kill a non existent thread"<<aslend;
		return -1;
	}
	
	if ((res = thr_kill(t->AslGetThread(), SIGUSR1))!=0)
		aslerr<<"ASL THREAD MANAGER ERROR: Can't kill thread"<<aslend;

	return res;
}

int AslThreadManager::AslGetThreadID(void)
{
	return thr_self();
}

// methods of class AslThread

AslThread::AslThread(AslThreadManager* m, int tid)
{
	if(asl_debug)
		aslio<<"ASL THREAD: Creating a new thread"<<aslend;	
		
	id = tid;
	attrib=m;
	runMode = 0;
}

AslThread::~AslThread(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD: Destroying thread object"<<aslend;
}

int AslThread::AslExecuteThread(asl_thread_func_t f, any_t arg)
{
	int res;

	if(asl_debug)
		aslio<<"ASL THREAD: Executing thread"<<aslend;

	if ((res = thr_create(NULL,attrib->AslGetStackSize(), f, arg, runMode , &thread)) != 0 )
		aslerr<<"ASL THREAD ERROR: Can't allocate thread"<<aslend;

	return res; 
}

void  AslThread::AslSuspendThread(void)
{
	int res;

	if(asl_debug)
		aslio<<"ASL THREAD: Stopping thread execution"<<aslend;

	if ((res=thr_suspend(thread)) != 0)
		aslerr<<"ASL THREAD ERROR: Can't stop thread"<<aslend;
}

void  AslThread::AslContinueThread(void)
{
	int res;

	if(asl_debug)
		aslio<<"ASL THREAD: Continuing thread execution"<<aslend;

	if ((res=thr_continue(thread)) != 0)
		aslerr<<"ASL THREAD ERROR: Can't continue thread"<<aslend;
}

void  AslThread::AslRunModeLWP(void)
{
	if(asl_debug)
		aslio<<"ASL THREAD: Detaching thread"<<aslend;

	runMode = THR_BOUND | THR_NEW_LWP ;
}

int AslThread::AslGetThreadAnswer(asl_any_t* r)
{
	int res;
	
	if (asl_debug)
		aslio<<"ASL THREAD: Getting thread answer"<<aslend;

	if ((res=thr_join(thread, 0, r)) != 0)
		aslerr<<"ASL THREAD ERROR: Can't get thread answer"<<aslend;

	return res;
}

asl_thread_t AslThread::AslGetThread(void)
{
	return thread;
}

int AslThread::AslGetThreadId(void)
{
	return id;
}

// methods of class AslMutex

AslMutex::AslMutex(void)
{
//	if(asl_debug)
//		aslio<<"ASL THREAD: Creating mutex variable"<<aslend;

//	if (mutex_init(&m, USYNC_THREAD, NULL)!=0)
//		aslerr<<"ASL THREAD ERROR: Can't initialize mutex variable"<<aslend;

	mutex_init(&m, USYNC_THREAD, NULL);
}

AslMutex::~AslMutex(void)
{
//	if(asl_debug)
//		aslio<<"ASL THREAD: Deleting mutex variable"<<aslend;

//	if (mutex_destroy(&m)!=0)
//		aslerr<<"ASL THREAD ERROR: Can't destroy mutex variable"<<aslend;

	mutex_destroy(&m);
}

asl_mutex_t *AslMutex::AslGetMutex(void)
{
	return &m;
}

int AslMutex::AslLock(void)
{
//	int res;

//	if(asl_debug)
//		aslio<<"ASL THREAD: Mutex Locking"<<aslend;

//	thr_yield();
//	if((res=mutex_lock(&m))!=0)
//		aslerr<<"ASL THREAD ERROR: Can't do mutex lock";

//	return res;
	return mutex_lock(&m);
}

int AslMutex::AslTryLock(void)
{
//	int res;

//	if(asl_debug)
//		aslio<<"ASL THREAD: Mutex try Locking"<<aslend;

//	thr_yield();
//	if((res=mutex_trylock(&m))!=0)
//		aslerr<<"ASL THREAD ERROR: Can't do mutex trylock"<<aslend;

//	return res;
	return mutex_trylock(&m);
}

int AslMutex::AslUnlock(void)
{
//	int res;

//	if(asl_debug)
//		aslio<<"ASL THREAD: Mutex Unlocking"<<aslend;

//	if((res=mutex_unlock(&m))!=0)
//		aslerr<<"ASL THREAD ERROR: Can't do mutex unlock"<<aslend;

	return mutex_unlock(&m);
}

// methods of class AslCondition

AslCondition::AslCondition(void)
{
//	if (asl_debug)
//		aslio<<"ASL THREAD: Creating condition variable"<<aslend;
	
//	if (cond_init(&c, USYNC_THREAD, NULL)!=0)
//		aslerr<<"ASL THREAD ERROR: Can't initialize mutex variable"<<aslend;	

	cond_init(&c, USYNC_THREAD, NULL);
}

AslCondition::~AslCondition(void)
{
//	if (asl_debug)
//		aslio<<"ASL THREAD: Deleting condition variable"<<aslend;
	
//	if (cond_destroy(&c)!=0)
//		aslerr<<"ASL THREAD ERROR: Can't delete condition variable"<<aslend;

	cond_destroy(&c);
}

int AslCondition::AslConditionWait(AslMutex* m)
{
//	if (asl_debug)
//		aslio<<"ASL THREAD: Waiting for the condition variable"<<aslend;
	
//	if (cond_wait(&c, m->AslGetMutex())!=0)
//	{
//		aslerr<<"ASL THREAD ERROR: Can't do wait condition"<<aslend;
//		return -1;
//	}

	return cond_wait(&c, m->AslGetMutex());
}

int AslCondition::AslConditionBroadcast(void)
{
//	if (asl_debug)
//		aslio<<"ASL THREAD: Waking up all threads that are waiting on the condition variable"<<aslend;

//	if (cond_broadcast(&c)!=0)
//	{
//		aslerr<<"ASL THREAD ERROR: Can't do broadcast condition"<<aslend;
//		return -1;
//	}

//	return 1;

	return cond_broadcast(&c);
}

asl_cond_t AslCondition::AslGetCondition(void)
{
	return c;
}

int AslCondition::AslConditionSignal(void)
{
//	if (asl_debug)
//		aslio<<"ASL THREAD: Waking up a thread that is waiting on the condition variable"<<aslend;

//	if(cond_signal(&c)!=0)
//	{
//		aslerr<<"ASL THREAD ERROR: Can't wake up threads"<<aslend;
//		return -1;
//	}

//	return 1;
	return cond_signal(&c);
}
